<copyright> tdbc manager class.
    Written by <a href="mailto:yacc@gerbil.org">Andreas Kostyrka</a>

    Copyright &copy; 1998 Andreas Kostyrka

    This file is part of tdbc.  tdbc is distributed under the terms of
    the GNU Library General Public License, a copy of which can be found
    in the tdbc distribution; see the file COPYING.LIB.

    <id>$Id: PGSQL.t,v 1.10 1999/12/22 22:58:19 tiggr Exp $</id>
    </copyright>

<c>
#include <tom/util.h>
#include <postgresql/libpq-fe.h>

DECL_USEL (tom, _pi__byteStringContents);
</c>

implementation class
PGSQLdriver : Driver, State

void
  load Array args
{
  [Manager addDriver [[self alloc] init]];
}

end;

implementation instance
PGSQLdriver

boolean
  urlOk String url
{
  if (["pgsql://" equal [url substring (0,8)]])
    = TRUE;
  else
    = FALSE;
}

<doc> A pgsql: url is composed as {pgsql://<hostname>:<port>/<dbname>}.
    </doc>
Connection
   connect String url
      name String name
  password String password
{
  // With this prefix, pgsql:/dbname or pgsql:dbname won't work...
  // Thu Aug 20 09:51:15 1998, tiggr@gerbil.org
  if (["pgsql://" equal [url substring (0,8)]])
    {
      String rest = [url substring (8,-1)];
      Indexed rest_comp = [rest componentsSeparatedBy '/']; 
      String hostfield = rest_comp[0];
      // Check bounds of array before indexing
      // Thu Aug 20 09:55:03 1998, tiggr@gerbil.org
      String dbname = rest_comp[1];
      MutableArray components = [hostfield componentsSeparatedBy ':'];
      String port;
      if ([components length] > 1)
	port = components[1];
      String host = components[0];
      if ([host length]==0)
	host = nil;
      = [[PGConnection alloc] init (dbname, host, port, name, password)];
    }
}

end;

implementation class
PGConnection: Connection, State

end;

implementation instance
PGConnection
{
  pointer pgdb;
  boolean intransaction;

  <doc> The result of the most recent command.  </doc>
  pointer pg_result;

  <doc> The index of the next row to be retrieved.  </doc>
  int currow;

  <doc> The number of rows that can be retrieved.  This is (of course) only
      valid after a SELECT query.  </doc>
  int num_rows;

  <doc> The number of columns retrieved.  </doc>
  int num_colums;
}

<doc> Designated initializer.  </doc>
id
  init (String, String, String, String, String)
       (dbname, host, port, user, passwd)
{
  pointer res;

  <c> {
    char *chost = 0, *cport = 0, *cuser = 0, *cpasswd = 0, *cdbname;
    int len;

    C_STRING_WITH_TOM_STRING (cdbname, len, dbname);
    if (host)
      C_STRING_WITH_TOM_STRING (chost, len, host);
    if (port)
      C_STRING_WITH_TOM_STRING (cport, len, port);
    if (user)
      C_STRING_WITH_TOM_STRING (cuser, len, user);
    if (passwd)
      C_STRING_WITH_TOM_STRING (cpasswd, len, passwd);

    res = PQsetdbLogin (chost, cport, 0, 0, cdbname, cuser, cpasswd);
    if (PQstatus ((PGconn *) res) != CONNECTION_OK)
      {
	PQfinish (res);
	res = 0;
      }
  } </c>

  if (!!res)
    {
      pgdb = res;
      = self;
    }
}

void
  close
{
  pointer cdb = pgdb;
  <c> if (cdb) PQfinish(cdb); cdb = 0; </c>
  pgdb = cdb;
}

void
  dealloc
{
  [self close];
}

void
  commit
{
  [self lowlevelSend "COMMIT;"];
  intransaction = FALSE;
}

void
  rollback
{
  [self lowlevelSend "ABORT;"];
  intransaction = FALSE;
}

protected void
  lowlevelSend String command
{
  pointer cdb = pgdb;
  pointer cres = pg_result;
  <c> {
    char *ccommand = 0;
    tom_int len;

    C_STRING_WITH_TOM_STRING (ccommand, len, command);
    if (cres)
      PQclear (cres);
      
    cres = PQexec (cdb, ccommand);
    if (!cres)
      fprintf (stderr, "lowlevelSend: %s: %s\n",
	       ccommand, PQerrorMessage ((PGconn *) cdb));
  } </c>
  pg_result = cres;
  num_rows = -1;
  currow = 0;
}

ObjectArray
  description
{
  = nil;
}

int
  execute String sqlcommand
{
  = 1;

  if (!intransaction)
    {
      [self lowlevelSend "BEGIN;"];
      intransaction = TRUE;
    }

  [self lowlevelSend sqlcommand];
}

MutableArray (result)
  fetchone
{
  if (num_rows < 0)
    [self num_rows];

  if (num_rows < 0 || currow >= num_rows)
    return nil;

  = [self fetchone [MutableObjectArray withCapacity num_colums]];
}

<doc> Retrieve a single row, re-using the {columns} array.  </doc>
MutableArray
  fetchone MutableArray columns
{
  if (num_rows < 0)
    [self num_rows];

  if (num_rows < 0 || currow >= num_rows)
    return nil;

  [columns truncate 0];

  pointer cres = pg_result;
  int nfields;
  <c> nfields = PQnfields (cres); </c>

  int i, cr = currow;
  String data;

  for (i = 0; i < nfields; i++)
    {
      <c> data = (void *) byte_string_with_c_string
				 (PQgetvalue ((PGresult *) cres, cr, i));
      </c>
      [columns add data];
  }

  currow = currow + 1;
  = columns;
}

ObjectArray 
  fetchmany int size
{
  MutableObjectArray res = [MutableObjectArray new];
  int i;

  for (i = 0; i < size; i++)
    {
      ObjectArray row = [self fetchone];
      if (row != nil)
	[res add row];
      else
	return res;
    }
  = res;
}

MutableObjectArray 
  fetchall
{
  MutableObjectArray res = [MutableObjectArray new];

  while (TRUE)
    {
      ObjectArray row = [self fetchone];
      if (row != nil) 
	[res add row];
      else
	return res;
    }
}

Cursor
  cursor
{
  = nil;
}

int
  num_rows
{
  if (num_rows == -1 && !!pg_result)
    {
      pointer cresult = pg_result;
      int n;

      <c> {
	n = PQntuples (cresult);
      } </c>

      num_rows = n;
    }

  = num_rows;
}

<doc> Perform the {query} and return the integer value of the single-column
    single-row numeric result.  Very handy, though maybe it should (a) not
    be here, or (b) also be in a normal {Connection}.  </doc>
int
  intQuery String query
{
  [self execute query];
  Indexed result;

  if ([self num_rows] == 1)
    {
      result = [self fetchone];
      if ([result length] == 1)
	return [result[0] intValue];
    }
  [[[stdio err] print (cmd, "WARNING: weird result (", result, ") for query ",
		       query)] nl];
}

end;
